import csv
import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_parsimony_pulp.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')


class TestExportTypes(TestCase):
    def test_export_types(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = pyproteininference.parameters.ProteinInferenceParameter(
            yaml_param_filepath=PARAMETER_FILE
        )

        self.assertEqual(protein_inference_parameters.digest_type, "trypsin")
        self.assertEqual(protein_inference_parameters.export, "q_value")
        self.assertEqual(protein_inference_parameters.fdr, 0.01)
        self.assertEqual(protein_inference_parameters.missed_cleavages, 3)
        self.assertEqual(protein_inference_parameters.picker, True)
        self.assertEqual(protein_inference_parameters.restrict_pep, 0.9)
        self.assertEqual(protein_inference_parameters.restrict_peptide_length, 7)
        self.assertEqual(protein_inference_parameters.restrict_q, 0.9)
        self.assertEqual(protein_inference_parameters.protein_score, "multiplicative_log")
        self.assertEqual(protein_inference_parameters.psm_score, "posterior_error_prob")
        self.assertEqual(protein_inference_parameters.psm_score_type, "multiplicative")
        self.assertEqual(protein_inference_parameters.decoy_symbol, "##")
        self.assertEqual(protein_inference_parameters.isoform_symbol, "-")
        self.assertEqual(protein_inference_parameters.reviewed_identifier_symbol, "sp|")
        self.assertEqual(protein_inference_parameters.inference_type, "parsimony")
        self.assertEqual(protein_inference_parameters.tag, "test_parsimony")
        self.assertEqual(protein_inference_parameters.grouping_type, "shared_peptides")
        self.assertEqual(protein_inference_parameters.max_identifiers_peptide_centric, 5)
        self.assertEqual(protein_inference_parameters.lp_solver, "pulp")

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = pyproteininference.in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data.restrict_psm_data()

        self.assertEqual(len(data.main_data_restricted), 26)

        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        data.create_scoring_input()

        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        if protein_inference_parameters.inference_type == pyproteininference.inference.Inference.EXCLUSION:
            # This gets ran if we run exclusion...
            data.exclude_non_distinguishing_peptides()

        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        score = pyproteininference.scoring.Score(data=data)
        score.score_psms(score_method=protein_inference_parameters.protein_score)

        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        if protein_inference_parameters.picker:
            data.protein_picker()
        else:
            pass

        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        inference_type = protein_inference_parameters.inference_type

        if inference_type == pyproteininference.inference.Inference.PARSIMONY:
            group = pyproteininference.inference.Parsimony(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.INCLUSION:
            group = pyproteininference.inference.Inclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.EXCLUSION:
            group = pyproteininference.inference.Exclusion(data=data, digest=digest)
            group.infer_proteins()

        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        data.calculate_q_values()

        export_type = "peptides"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have peptides space separated
        self.assertListEqual(
            output[1],
            [
                "RPOC_SHIF8|Q0SY12",
                "82.89306334778564",
                "0.0",
                "12",
                "Reviewed",
                "1",
                "CGVEVTQTK EGLNVLQY#FISTHGAR FATSDLNDLYR IALASPDMIR IPQESGGTK LIPAGTGYAYHQDR "
                "MGAEAIQALLK NTLLHEQWCDLLEENSVDAVK RVDYSGR VADLFEAR VIDIWAAANDR VTAEDVLKPGTADILVPR",
            ],
        )

        export_type = "psms"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have psms space separated
        self.assertListEqual(
            output[1],
            [
                "RPOC_SHIF8|Q0SY12",
                "82.89306334778564",
                "0.0",
                "12",
                "Reviewed",
                "1",
                "CGVEVTQTK EGLNVLQY#FISTHGAR FATSDLNDLYR IALASPDMIR IPQESGGTK LIPAGTGYAYHQDR "
                "MGAEAIQALLK NTLLHEQWCDLLEENSVDAVK RVDYSGR VADLFEAR VIDIWAAANDR VTAEDVLKPGTADILVPR",
            ],
        )

        export_type = "psm_ids"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have psm_ids space separated
        self.assertListEqual(
            output[1],
            [
                "RPOC_SHIF8|Q0SY12",
                "82.89306334778564",
                "0.0",
                "12",
                "Reviewed",
                "1",
                "13 14 15 16 17 18 19 20 21 22 23 24",
            ],
        )

        export_type = "q_value"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have peptides comma separated
        self.assertListEqual(
            output[1],
            [
                "RPOC_SHIF8|Q0SY12",
                "82.89306334778564",
                "0.0",
                "12",
                "Reviewed",
                "1",
                "CGVEVTQTK",
                "EGLNVLQY#FISTHGAR",
                "FATSDLNDLYR",
                "IALASPDMIR",
                "IPQESGGTK",
                "LIPAGTGYAYHQDR",
                "MGAEAIQALLK",
                "NTLLHEQWCDLLEENSVDAVK",
                "RVDYSGR",
                "VADLFEAR",
                "VIDIWAAANDR",
                "VTAEDVLKPGTADILVPR",
            ],
        )

        export_type = "q_value_all"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have peptides comma separated
        self.assertListEqual(
            output[1],
            [
                "RPOC_SHIF8|Q0SY12",
                "82.89306334778564",
                "0.0",
                "12",
                "Reviewed",
                "1",
                "CGVEVTQTK",
                "EGLNVLQY#FISTHGAR",
                "FATSDLNDLYR",
                "IALASPDMIR",
                "IPQESGGTK",
                "LIPAGTGYAYHQDR",
                "MGAEAIQALLK",
                "NTLLHEQWCDLLEENSVDAVK",
                "RVDYSGR",
                "VADLFEAR",
                "VIDIWAAANDR",
                "VTAEDVLKPGTADILVPR",
            ],
        )

        export_type = "q_value_comma_sep"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have proteins comma separated
        self.assertListEqual(
            output[2],
            [
                "RAF1_HUMAN|P04049",
                "70.7434325345954",
                "0.0",
                "6",
                "Reviewed",
                "2",
                "ARAF_HUMAN|P10398",
                "BRAF_HUMAN|P15056",
            ],
        )

        export_type = "leads"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have peptides comma separated with no q values
        self.assertListEqual(
            output[2],
            [
                "RAF1_HUMAN|P04049",
                "70.7434325345954",
                "6",
                "Reviewed",
                "{2, 3}",
                "CQTCGYKFHEHCSTK",
                "FQMFQLIDIAR",
                "QTAQGMDYLHAK",
                "SASEPSLHR",
                "VFLPNKQR",
                "WHGDVAVKILK",
            ],
        )

        export_type = "all"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have peptides comma separated with no q values
        self.assertListEqual(
            output[-1],
            [
                "Q96BA7_HUMAN|Q96BA7",
                "6.907755278982137",
                "1",
                "Unreviewed",
                "{5}",
                "LQAALDDEEAGGRPAMEPGNGSLDLGGDSAGR",
            ],
        )

        export_type = "comma_sep"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should have proteins comma separated with no q values
        self.assertListEqual(
            output[-1],
            [
                "HNRPU_HUMAN|Q00839",
                "15.316094065486292",
                "2",
                "Reviewed",
                "{5}",
                "B3KX72_HUMAN|B3KX72",
                "Q96BA7_HUMAN|Q96BA7",
            ],
        )

        export_type = "long"
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=OUTPUT_DIR, export_type=export_type)

        output = []
        with open(export.filepath, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                output.append(row)

        # Output should be one line for every peptide
        self.assertListEqual(
            output[1],
            ['RPOC_SHIF8|Q0SY12', '82.89306334778564', '0.0', '12', 'Reviewed', '1', 'CGVEVTQTK'],
        )

        self.assertListEqual(
            output[2],
            ['RPOC_SHIF8|Q0SY12', '82.89306334778564', '0.0', '12', 'Reviewed', '1', 'EGLNVLQY#FISTHGAR'],
        )

        self.assertListEqual(
            output[-1],
            [
                '##TCAF2_HUMAN|##A6NFQ2',
                '2.4079456086518722',
                '0.3333333333333333',
                '1',
                'Reviewed',
                '6',
                'MEPTPVPFCGAK',
            ],
        )



import logging
import sys
from collections import OrderedDict

import pulp

from pyproteininference import datastore
from pyproteininference.physical import ProteinGroup, Psm

logger = logging.getLogger(__name__)

# set up our logger
logging.basicConfig(
    stream=sys.stderr,
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class Inference(object):
    """
    Parent Inference class for all inference/grouper subset classes.
    The base Inference class contains several methods that are shared across the Inference sub-classes.

    Attributes:
        data (DataStore): [DataStore object][pyproteininference.datastore.DataStore].
        digest (Digest): [Digest object][pyproteininference.in_silico_digest.Digest].
    """

    PARSIMONY = "parsimony"
    INCLUSION = "inclusion"
    EXCLUSION = "exclusion"
    FIRST_PROTEIN = "first_protein"
    PEPTIDE_CENTRIC = "peptide_centric"

    INFERENCE_TYPES = [
        PARSIMONY,
        INCLUSION,
        EXCLUSION,
        FIRST_PROTEIN,
        PEPTIDE_CENTRIC,
    ]

    INFERENCE_NAME_MAP = {
        PARSIMONY: "Parsimony",
        INCLUSION: "Inclusion",
        EXCLUSION: "Exclusion",
        FIRST_PROTEIN: "First Protein",
        PEPTIDE_CENTRIC: "Peptide Centric",
    }

    SUBSET_PEPTIDES = "subset_peptides"
    SHARED_PEPTIDES = "shared_peptides"
    NONE_GROUPING = None

    GROUPING_TYPES = [SUBSET_PEPTIDES, SHARED_PEPTIDES, NONE_GROUPING]

    PULP = "pulp"
    LP_SOLVERS = [PULP]

    ALL_SHARED_PEPTIDES = "all"
    BEST_SHARED_PEPTIDES = "best"
    NONE_SHARED_PEPTIDES = None
    SHARED_PEPTIDE_TYPES = [
        ALL_SHARED_PEPTIDES,
        BEST_SHARED_PEPTIDES,
        NONE_SHARED_PEPTIDES,
    ]

    def __init__(self, data, digest):
        """
        Initialization method of Inference object.

        Args:
            data (DataStore): [DataStore object][pyproteininference.datastore.DataStore].
            digest (Digest): [Digest object][pyproteininference.in_silico_digest.Digest].

        """
        self.data = data
        self.digest = digest
        self.data._validate_scored_proteins()

    @classmethod
    def run_inference(cls, data, digest):
        """
        This class method dispatches to one of the five different inference classes/models
        based on input from the [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter]
        object.
        The methods are "parsimony", "inclusion", "exclusion", "peptide_centric", and "first_protein".

        Args:
            data (DataStore): [DataStore object][pyproteininference.datastore.DataStore].
            digest (Digest): [Digest object][pyproteininference.in_silico_digest.Digest].

        Example:
            >>> pyproteininference.inference.Inference.run_inference(data=data,digest=digest)

        """

        inference_type = data.parameter_file_object.inference_type

        logger.info("Running Inference with Inference Type: {}".format(inference_type))

        if inference_type == Inference.PARSIMONY:
            group = Parsimony(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == Inference.INCLUSION:
            group = Inclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == Inference.EXCLUSION:
            group = Exclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == Inference.FIRST_PROTEIN:
            group = FirstProtein(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == Inference.PEPTIDE_CENTRIC:
            group = PeptideCentric(data=data, digest=digest)
            group.infer_proteins()

    def _create_protein_groups(self, scored_proteins):
        """
        This method sets up protein groups for inference methods that do not need grouping.

        Args:
            scored_proteins (list): List of scored [Protein][pyproteininference.physical.Protein] objects.

        Returns:
            list: List of lists of scored [Protein][pyproteininference.physical.Protein] objects.

        """
        scored_proteins = sorted(
            scored_proteins,
            key=lambda k: (k.score, len(k.raw_peptides), k.identifier),
            reverse=True,
        )

        prot_pep_dict = self.data.protein_to_peptide_dictionary()

        restricted_peptides_set = set(self.data.restricted_peptides)

        grouped_proteins = []
        for protein_objects in scored_proteins:
            cur_protein_identifier = protein_objects.identifier

            # Set peptide variable if the peptide is in the restricted peptide set
            # Sort the peptides alphabetically
            protein_objects.peptides = set(
                sorted([x for x in prot_pep_dict[cur_protein_identifier] if x in restricted_peptides_set])
            )
            protein_list_group = [protein_objects]
            grouped_proteins.append(protein_list_group)
        return grouped_proteins

    def _apply_protein_group_ids(self, grouped_protein_objects):
        """
        This method creates the ProteinGroup objects from the output of
            [_create_protein_groups][`pyproteininference.inference.Inference._create_protein_groups].

        Args:
            grouped_protein_objects (list): list of grouped [Protein][pyproteininference.physical.Protein] objects.

        Returns:
            dict: a Dictionary that contains a list of [ProteinGroup][pyproteininference.physical.ProteinGroup]
                objects (key:"group_objects") and a list of grouped [Protein][pyproteininference.physical.Protein]
                objects (key:"grouped_protein_objects").


        """

        sp_protein_set = set(self.digest.swiss_prot_protein_set)

        prot_pep_dict = self.data.protein_to_peptide_dictionary()

        # Here we create group ID's
        group_id = 0
        protein_group_objects = []
        for protein_group in grouped_protein_objects:
            protein_list = []
            group_id = group_id + 1
            pg = ProteinGroup(group_id)
            logger.debug("Created Protein Group with ID: {}".format(str(group_id)))
            for protein in protein_group:
                cur_protein = protein
                # The following loop assigns group_id's, reviewed/unreviewed status, and number of unique peptides...
                if group_id not in cur_protein.group_identification:
                    cur_protein.group_identification.add(group_id)
                if protein.identifier in sp_protein_set:
                    cur_protein.reviewed = True
                else:
                    cur_protein.unreviewed = True
                cur_identifier = protein.identifier
                cur_protein.num_peptides = len(prot_pep_dict[cur_identifier])
                # Here append the number of unique peptides... so we can use this as secondary sorting...
                protein_list.append(cur_protein)
                # Sorted protein_groups then becomes a list of lists... of protein objects

            pg.proteins = protein_list
            protein_group_objects.append(pg)

        return_dict = {
            "grouped_protein_objects": grouped_protein_objects,
            "group_objects": protein_group_objects,
        }

        return return_dict


class Inclusion(Inference):
    """
    Inclusion Inference class. This class contains methods that support the initialization of an
    Inclusion inference method.

    Attributes:
        data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
        digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
        scored_data (list): a List of scored [Protein][pyproteininference.physical.Protein] objects.

    """

    def __init__(self, data, digest):
        """
        Initialization method of the Inclusion Inference method.

        Args:
            data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
        """

        self.data = data
        self.digest = digest
        self.data._validate_scored_proteins()
        self.scored_data = self.data.get_protein_data()

    def infer_proteins(self):
        """
        This method performs the grouping for Inclusion.

        Inclusion actually does not do grouping as all peptides get assigned to all possible proteins
        and groups are not created.

        This method assigns the variables: `grouped_scored_proteins` and `protein_group_objects`.
        These are both variables of the [DataStore Object][pyproteininference.datastore.DataStore] and are
        lists of [Protein][pyproteininference.physical.Protein] objects
        and [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.
        """

        grouped_proteins = self._create_protein_groups(scored_proteins=self.scored_data)

        hl = self.data.higher_or_lower()

        logger.info("Applying Group ID's for the Inclusion Method")

        regrouped_proteins = self._apply_protein_group_ids(
            grouped_protein_objects=grouped_proteins,
        )

        grouped_protein_objects = regrouped_proteins["grouped_protein_objects"]
        protein_group_objects = regrouped_proteins["group_objects"]

        logger.info("Sorting Results based on lead Protein Score")
        grouped_protein_objects = datastore.DataStore.sort_protein_objects(
            grouped_protein_objects=grouped_protein_objects, higher_or_lower=hl
        )
        protein_group_objects = datastore.DataStore.sort_protein_group_objects(
            protein_group_objects=protein_group_objects, higher_or_lower=hl
        )

        self.data.grouped_scored_proteins = grouped_protein_objects
        self.data.protein_group_objects = protein_group_objects

    def _apply_protein_group_ids(self, grouped_protein_objects):
        """
        This method creates the ProteinGroup objects for the inclusion inference type using protein groups from
         [_create_protein_groups][`pyproteininference.inference.Inference._create_protein_groups].

        Args:
            grouped_protein_objects (list): list of grouped [Protein][pyproteininference.physical.Protein] objects.

        Returns:
            dict: a Dictionary that contains a list of [ProteinGroup][pyproteininference.physical.ProteinGroup]
                objects (key:"group_objects") and a list of
                grouped [Protein][pyproteininference.physical.Protein] objects (key:"grouped_protein_objects").

        """

        sp_protein_set = set(self.digest.swiss_prot_protein_set)

        prot_pep_dict = self.data.protein_to_peptide_dictionary()

        # Here we create group ID's
        group_id = 0
        protein_group_objects = []
        for protein_group in grouped_protein_objects:
            protein_list = []
            group_id = group_id + 1
            pg = ProteinGroup(group_id)
            logger.debug("Created Protein Group with ID: {}".format(str(group_id)))
            for prot in protein_group:
                cur_protein = prot
                # The following loop assigns group_id's, reviewed/unreviewed status, and number of unique peptides...
                if group_id not in cur_protein.group_identification:
                    cur_protein.group_identification.add(group_id)
                if cur_protein.identifier in sp_protein_set:
                    cur_protein.reviewed = True
                else:
                    cur_protein.unreviewed = True
                cur_identifier = cur_protein.identifier
                cur_protein.num_peptides = len(prot_pep_dict[cur_identifier])
                # Here append the number of unique peptides... so we can use this as secondary sorting...
                protein_list.append(cur_protein)
                # Sorted protein_groups then becomes a list of lists... of protein objects

            pg.proteins = protein_list
            protein_group_objects.append(pg)

        return_dict = {
            "grouped_protein_objects": grouped_protein_objects,
            "group_objects": protein_group_objects,
        }

        return return_dict


class Exclusion(Inference):
    """
    Exclusion Inference class. This class contains methods that support the initialization of an
    Exclusion inference method.

    Attributes:
        data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
        digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
        scored_data (list): a List of scored [Protein][pyproteininference.physical.Protein] objects.

    """

    def __init__(self, data, digest):
        """
        Initialization method of the Exclusion Class.

        Args:
            data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].

        """
        self.data = data
        self.digest = digest
        self.data._validate_scored_proteins()
        self.scored_data = self.data.get_protein_data()
        self.list_of_prots_not_in_db = None
        self.list_of_peps_not_in_db = None

    def infer_proteins(self):
        """
        This method performs the Exclusion inference/grouping method.

        For the exclusion inference method groups cannot be created because all shared peptides are removed.

        This method assigns the variables: `grouped_scored_proteins` and `protein_group_objects`.
        These are both variables of the [DataStore Object][pyproteininference.datastore.DataStore] and are
        lists of [Protein][pyproteininference.physical.Protein] objects
        and [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        """

        grouped_proteins = self._create_protein_groups(scored_proteins=self.scored_data)

        hl = self.data.higher_or_lower()

        logger.info("Applying Group ID's for the Exclusion Method")
        regrouped_proteins = self._apply_protein_group_ids(
            grouped_protein_objects=grouped_proteins,
        )

        grouped_protein_objects = regrouped_proteins["grouped_protein_objects"]
        protein_group_objects = regrouped_proteins["group_objects"]

        logger.info("Sorting Results based on lead Protein Score")
        grouped_protein_objects = datastore.DataStore.sort_protein_objects(
            grouped_protein_objects=grouped_protein_objects, higher_or_lower=hl
        )
        protein_group_objects = datastore.DataStore.sort_protein_group_objects(
            protein_group_objects=protein_group_objects, higher_or_lower=hl
        )

        self.data.grouped_scored_proteins = grouped_protein_objects
        self.data.protein_group_objects = protein_group_objects


class Parsimony(Inference):
    """
    Parsimony Inference class. This class contains methods that support the initialization of a
    Parsimony inference method.

    Attributes:
        data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
        digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
        scored_data (list): a List of scored [Protein][pyproteininference.physical.Protein] objects.
        lead_protein_set (set): Set of protein strings that are classified as leads from the LP solver.

    """

    def __init__(self, data, digest):
        """
        Initialization method of the Parsimony object.

        Args:
            data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
        """
        self.data = data
        self.digest = digest
        self.data._validate_scored_proteins()
        self.scored_data = self.data.get_protein_data()
        self.lead_protein_set = None
        self.parameter_file_object = data.parameter_file_object

    def _create_protein_groups(
        self,
        all_scored_proteins,
        lead_protein_objects,
        grouping_type="shared_peptides",
    ):
        """
        Internal method that creates a list of lists of [Protein][pyproteininference.physical.Protein]
        objects for the Parsimony inference object.
        These list of lists are "groups" and the proteins get grouped them according to grouping_type variable.

        Args:
            all_scored_proteins (list): list of [Protein][pyproteininference.physical.Protein] objects.
            lead_protein_objects (list): list of [Protein][pyproteininference.physical.Protein] objects
                Only needed if inference_type=parsimony.
            grouping_type: (str): One of `GROUPING_TYPES`.

        Returns:
            list: list of lists of [Protein][pyproteininference.physical.Protein] objects.

        """

        logger.info("Grouping Peptides with Grouping Type: {}".format(grouping_type))
        logger.info("Grouping Peptides with Inference Type: {}".format(self.PARSIMONY))

        all_scored_proteins = sorted(
            all_scored_proteins,
            key=lambda k: (len(k.raw_peptides), k.identifier),
            reverse=True,
        )

        lead_scored_proteins = lead_protein_objects
        lead_scored_proteins = sorted(
            lead_scored_proteins,
            key=lambda k: (len(k.raw_peptides), k.identifier),
            reverse=True,
        )

        protein_finder = [x.identifier for x in all_scored_proteins]

        prot_pep_dict = self.data.protein_to_peptide_dictionary()

        protein_tracker = set()
        restricted_peptides_set = set(self.data.restricted_peptides)
        try:
            picked_removed = set([x.identifier for x in self.data.picked_proteins_removed])
        except TypeError:
            picked_removed = set()

        missing_proteins = set()
        in_silico_peptides_to_proteins = self.digest.peptide_to_protein_dictionary
        grouped_proteins = []
        for protein_objects in lead_scored_proteins:
            if protein_objects not in protein_tracker:
                protein_tracker.add(protein_objects)
                cur_protein_identifier = protein_objects.identifier

                # Set peptide variable if the peptide is in the restricted peptide set
                # Sort the peptides alphabetically
                protein_objects.peptides = set(
                    sorted([x for x in prot_pep_dict[cur_protein_identifier] if x in restricted_peptides_set])
                )
                protein_list_group = [protein_objects]
                current_peptides = prot_pep_dict[cur_protein_identifier]

                current_grouped_proteins = set()
                for (
                    peptide
                ) in current_peptides:  # Probably put an if here... if peptide is in the list of peptide after being
                    # restricted by datastore.RestrictMainData
                    if peptide in restricted_peptides_set:
                        # Get the proteins that map to the current peptide using in_silico_peptides_to_proteins
                        # First make sure our peptide is formatted properly...
                        if not peptide.isupper() or not peptide.isalpha():
                            # If the peptide is not all upper case or if its not all alphabetical...
                            peptide = Psm.remove_peptide_mods(peptide)
                        potential_protein_list = in_silico_peptides_to_proteins[peptide]
                        if not potential_protein_list:
                            logger.warning(
                                "Protein {} and Peptide {} is not in database...".format(
                                    protein_objects.identifier, peptide
                                )
                            )

                        # Assign proteins to groups based on shared peptide... unless the protein is equivalent
                        # to the current identifier
                        if grouping_type != self.NONE_GROUPING:
                            for protein in potential_protein_list:
                                # If statement below to avoid grouping the same protein twice and to not group the lead
                                if (
                                    protein not in current_grouped_proteins
                                    and protein != cur_protein_identifier
                                    and protein not in picked_removed
                                    and protein not in missing_proteins
                                ):
                                    try:
                                        # Try to find its object using protein_finder (list of identifiers) and
                                        # lead_scored_proteins (list of Protein Objects)
                                        cur_index = protein_finder.index(protein)
                                        current_protein_object = all_scored_proteins[cur_index]
                                        if not current_protein_object.peptides:
                                            current_protein_object.peptides = set(
                                                sorted(
                                                    [
                                                        x
                                                        for x in prot_pep_dict[current_protein_object.identifier]
                                                        if x in restricted_peptides_set
                                                    ]
                                                )
                                            )
                                        if grouping_type == self.SHARED_PEPTIDES:
                                            current_grouped_proteins.add(current_protein_object)
                                        elif grouping_type == self.SUBSET_PEPTIDES:
                                            if current_protein_object.peptides.issubset(protein_objects.peptides):
                                                current_grouped_proteins.add(current_protein_object)
                                                protein_tracker.add(current_protein_object)
                                            else:
                                                pass
                                        else:
                                            pass
                                    except ValueError:
                                        logger.warning(
                                            "Protein from DB {} not found with protein finder for peptide {}".format(
                                                protein, peptide
                                            )
                                        )
                                        missing_proteins.add(protein)

                                else:
                                    pass
                # Add the proteins to the lead if they share peptides...
                protein_list_group = protein_list_group + list(current_grouped_proteins)
                # protein_list_group at first is just the lead protein object...
                # We then try apply grouping by looking at all peptide from the lead...
                # For all of these peptide look at all other non lead proteins and try to assign them to the group...
                # We assign the entire protein object as well... in the above try/except
                # Then append this sub group to the main list
                # The variable grouped_proteins is now a list of lists which each element being a Protein object and
                # each list of protein objects corresponding to a group
                grouped_proteins.append(protein_list_group)

        return grouped_proteins

    def _swissprot_and_isoform_override(
        self,
        scored_data,
        grouped_proteins,
        override_type="soft",
        isoform_override=True,
    ):
        """
        This internal method creates and reorders protein groups based on criteria such as Reviewed/Unreviewed
        Identifiers as well as Canonincal/Isoform Identifiers.
        This method is only used with parsimony inference type.

        Args:
            scored_data (list): list of scored [Protein][pyproteininference.physical.Protein] objects.
            grouped_proteins:  list of grouped [Protein][pyproteininference.physical.Protein] objects.
            override_type (str): "soft" or "hard" to indicate Reviewed/Unreviewed override. "soft" is preferred and
                default.
            isoform_override (bool): True/False on whether to favor canonical forms vs isoforms as group leads.

        Returns:
            dict: a Dictionary that contains a list of [ProteinGroup][pyproteininference.physical.ProteinGroup] objects
            (key:"group_objects") and a list of grouped [Protein][pyproteininference.physical.Protein]
            objects (key:"grouped_protein_objects").


        """

        sp_protein_set = set(self.digest.swiss_prot_protein_set)
        scored_proteins = list(scored_data)
        protein_finder = [x.identifier for x in scored_proteins]

        prot_pep_dict = self.data.protein_to_peptide_dictionary()

        # Get the higher or lower variable
        higher_or_lower = self.data.higher_or_lower()

        logger.info("Applying Group IDs... and Executing {} Swissprot Override...".format(override_type))
        # Here we create group ID's for all groups and do some sorting
        grouped_protein_objects = []
        group_id = 0
        leads = set()
        protein_group_objects = []
        for protein_group in grouped_proteins:
            protein_list = []
            group_id = group_id + 1
            # Make a protein group
            pg = ProteinGroup(group_id)
            logger.debug("Created Protein Group with ID: {}".format(str(group_id)))
            for prots in protein_group:
                # Loop over all proteins in the original group
                try:
                    # The following loop assigns group_id's, reviewed/unreviewed status, and number of unique peptides
                    pindex = protein_finder.index(prots.identifier)
                    # Attempt to find the protein object by identifier
                    cur_protein = scored_proteins[pindex]
                    if group_id not in cur_protein.group_identification:
                        cur_protein.group_identification.add(group_id)
                    if prots.identifier in sp_protein_set:
                        cur_protein.reviewed = True
                    else:
                        cur_protein.unreviewed = True
                    cur_identifier = prots.identifier
                    cur_protein.num_peptides = len(prot_pep_dict[cur_identifier])
                    # Here append the number of unique peptides... so we can use this as secondary sorting...
                    protein_list.append(cur_protein)
                    # Sorted groups then becomes a list of lists... of protein objects

                except ValueError:
                    # Here we pass if the protein does not have a score...
                    # Potentially it got 'picked' (removed) by protein picker...
                    pass

            # Sort protein sub group
            protein_list = datastore.DataStore.sort_protein_sub_groups(
                protein_list=protein_list, higher_or_lower=higher_or_lower
            )

            # grouped_protein_objects is the MAIN list of lists with grouped protein objects
            grouped_protein_objects.append(protein_list)
            # If the lead is reviewed append it to leads and do nothing else...
            # If the lead is unreviewed then try to replace it with the best reviewed hit
            # Run swissprot override
            if self.data.parameter_file_object.reviewed_identifier_symbol:
                sp_override = self._swissprot_override(
                    protein_list=protein_list,
                    leads=leads,
                    grouped_protein_objects=grouped_protein_objects,
                    override_type=override_type,
                )
                grouped_protein_objects = sp_override["grouped_protein_objects"]
                leads = sp_override["leads"]
                protein_list = sp_override["protein_list"]

            # Run isoform override If we want to run isoform_override and if the isoform symbol exists...
            if isoform_override and self.data.parameter_file_object.isoform_symbol:
                iso_override = self._isoform_override(
                    protein_list=protein_list,
                    leads=leads,
                    grouped_protein_objects=grouped_protein_objects,
                )
                grouped_protein_objects = iso_override["grouped_protein_objects"]
                leads = iso_override["leads"]
                protein_list = iso_override["protein_list"]

            pg.proteins = protein_list
            protein_group_objects.append(pg)

        return_dict = {
            "grouped_protein_objects": grouped_protein_objects,
            "group_objects": protein_group_objects,
        }

        return return_dict

    def _swissprot_override(self, protein_list, leads, grouped_protein_objects, override_type):
        """
        This method re-assigns protein group leads if the lead is an unreviewed protein and if the protein group
         contains a reviewed protein that contains the exact same set of peptides as the unreviewed lead.
        This method is here to provide consistency to the output.

        Args:
            protein_list (list): List of grouped [Protein][pyproteininference.physical.Protein] objects.
            leads (set): Set of string protein identifiers that have been identified as a lead.
            grouped_protein_objects (list): List of protein_list lists.
            override_type (str): "soft" or "hard" on how to override non reviewed identifiers. "soft" is preferred.

        Returns:
            dict: leads (set): Set of string protein identifiers that have been identified as a lead.
             Updated to reflect lead changes.
            grouped_protein_objects (list): List of protein_list lists. Updated to reflect lead changes.
            protein_list (list): List of grouped [Protein][pyproteininference.physical.Protein] objects.
                Updated to reflect lead changes.

        """

        if not protein_list[0].reviewed:
            # If the lead is unreviewed attempt to replace it...
            # Start to loop through protein_list which is the current group...
            for protein in protein_list[1:]:
                # Find the first reviewed it... if its not a lead protein already then do score swap and break...
                if protein.reviewed:
                    best_swiss_prot_prot = protein

                    if override_type == "soft":
                        # If the lead proteins peptides are a subset of the best swissprot.... then swap the proteins.
                        # (meaning equal peptides or the swissprot completely covers the tremble reference)
                        if best_swiss_prot_prot.identifier not in leads and set(protein_list[0].peptides).issubset(
                            set(best_swiss_prot_prot.peptides)
                        ):
                            # We use -1 as the idex of grouped_protein_objects because the current 'protein_list' is
                            # the last entry appended to scores grouped
                            # Essentially grouped_protein_objects[-1]==protein_list
                            # We need this syntax so we can switch the location of the unreviewed lead identifier with
                            # the best reviewed identifier in grouped_protein_objects
                            swiss_prot_override_index = grouped_protein_objects[-1].index(best_swiss_prot_prot)
                            cur_tr_lead = grouped_protein_objects[-1][0]
                            (
                                grouped_protein_objects[-1][0],
                                grouped_protein_objects[-1][swiss_prot_override_index],
                            ) = (
                                grouped_protein_objects[-1][swiss_prot_override_index],
                                grouped_protein_objects[-1][0],
                            )
                            grouped_protein_objects[-1][swiss_prot_override_index], grouped_protein_objects[-1][0]
                            new_sp_lead = grouped_protein_objects[-1][0]
                            logger.info(
                                "Overriding Unreviewed {} with Reviewed {}".format(
                                    cur_tr_lead.identifier, new_sp_lead.identifier
                                )
                            )

                            # Append new_sp_lead protein to leads, to make sure we dont repeat leads
                            leads.add(new_sp_lead.identifier)
                            break
                        else:
                            # If no reviewed and none not in leads then pass...
                            pass

                    if override_type == "hard":
                        if best_swiss_prot_prot.identifier not in leads:
                            # We use -1 as the index of grouped_protein_objects because the current 'protein_list'
                            # is the last entry appended to grouped_protein_objects
                            # Essentially grouped_protein_objects[-1]==protein_list
                            # We need this syntax so we can switch the location of the unreviewed lead identifier
                            # with the best reviewed identifier in grouped_protein_objects
                            swiss_prot_override_index = grouped_protein_objects[-1].index(best_swiss_prot_prot)
                            cur_tr_lead = grouped_protein_objects[-1][0]
                            # Re-assigning the value within the index will also reassign the value in protein_list...
                            # This is because grouped_protein_objects[-1] equals protein_list
                            # So we do not have to reassign values in protein_list
                            (
                                grouped_protein_objects[-1][0],
                                grouped_protein_objects[-1][swiss_prot_override_index],
                            ) = (
                                grouped_protein_objects[-1][swiss_prot_override_index],
                                grouped_protein_objects[-1][0],
                            )
                            new_sp_lead = grouped_protein_objects[-1][0]
                            logger.info(
                                "Overriding Unreviewed {} with Reviewed {}".format(
                                    cur_tr_lead.identifier, new_sp_lead.identifier
                                )
                            )

                            # Append new_sp_lead protein to leads, to make sure we dont repeat leads
                            leads.add(new_sp_lead.identifier)
                            break
                        else:
                            # If no reviewed and none not in leads then pass...
                            pass

                else:
                    pass

        else:
            leads.add(protein_list[0].identifier)

        return_dict = {
            "leads": leads,
            "grouped_protein_objects": grouped_protein_objects,
            "protein_list": protein_list,
        }

        return return_dict

    def _isoform_override(self, protein_list, grouped_protein_objects, leads):
        """
        This method re-assigns protein group leads if the lead is an isoform protein and if the protein group contains
        a canonical protein that contains the exact same set of peptides as the isoform lead.
        This method is here to provide consistency to the output.

        Args:
            protein_list (list): List of grouped [Protein][pyproteininference.physical.Protein] objects.
            leads (set): Set of string protein identifiers that have been identified as a lead.
            grouped_protein_objects (list): List of protein_list lists.

        Returns:
            dict: leads (set): Set of string protein identifiers that have been identified as a lead. Updated to
                reflect lead changes.
            grouped_protein_objects (list): List of protein_list lists. Updated to reflect lead changes.
            protein_list (list): List of grouped [Protein][pyproteininference.physical.Protein] objects.
                Updated to reflect lead changes.


        """

        if self.data.parameter_file_object.isoform_symbol in protein_list[0].identifier:
            pure_id = protein_list[0].identifier.split(self.data.parameter_file_object.isoform_symbol)[0]
            # Start to loop through protein_list which is the current group...
            for potential_replacement in protein_list[1:]:
                isoform_override = potential_replacement
                if (
                    isoform_override.identifier == pure_id
                    and isoform_override.identifier not in leads
                    and set(protein_list[0].peptides).issubset(set(isoform_override.peptides))
                ):
                    isoform_override_index = grouped_protein_objects[-1].index(isoform_override)
                    cur_iso_lead = grouped_protein_objects[-1][0]
                    # Re-assigning the value within the index will also reassign the value in protein_list...
                    # This is because grouped_protein_objects[-1] equals protein_list
                    # So we do not have to reassign values in protein_list
                    (grouped_protein_objects[-1][0], grouped_protein_objects[-1][isoform_override_index],) = (
                        grouped_protein_objects[-1][isoform_override_index],
                        grouped_protein_objects[-1][0],
                    )
                    grouped_protein_objects[-1][isoform_override_index], grouped_protein_objects[-1][0]

                    new_iso_lead = grouped_protein_objects[-1][0]
                    logger.info(
                        "Overriding Isoform {} with {}".format(cur_iso_lead.identifier, new_iso_lead.identifier)
                    )
                    leads.add(protein_list[0].identifier)

        return_dict = {
            "leads": leads,
            "grouped_protein_objects": grouped_protein_objects,
            "protein_list": protein_list,
        }

        return return_dict

    def _reassign_protein_group_leads(self, protein_group_objects):
        """
        This internal method corrects leads that are improperly assigned in the parsimony inference method.
        This method acts on the protein group objects.

        Args:
            protein_group_objects (list): List of [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        Returns:
            protein_group_objects (list): List of [ProteinGroup][pyproteininference.physical.ProteinGroup] objects
            where leads have been reassigned properly.


        """

        # Get the higher or lower variable
        if not self.data.high_low_better:
            higher_or_lower = self.data.higher_or_lower()
        else:
            higher_or_lower = self.data.high_low_better

        # Sometimes we have cases where:
        # protein a maps to peptides 1,2,3
        # protein b maps to peptides 1,2
        # protein c maps to a bunch of peptides and peptide 3
        # Therefore, in the model proteins a and b are equivalent in that they map to 2 peptides together - 1 and 2.
        # peptide 3 maps to a but also to c...
        # Sometimes the model (pulp) will spit out protein b as the lead... we wish to swap protein b as the lead with
        # protein a because it will likely have a better score...
        logger.info("Potentially Reassigning Protein Group leads...")
        lead_protein_set = set([x.proteins[0].identifier for x in protein_group_objects])
        for i in range(len(protein_group_objects)):
            for j in range(1, len(protein_group_objects[i].proteins)):  # Loop over all sub proteins in the group...
                # if the lead proteins peptides are a subset of one of its proteins in the group, and the secondary
                # protein is not a lead protein and its score is better than the leads... and it has more peptides...
                new_lead = protein_group_objects[i].proteins[j]
                old_lead = protein_group_objects[i].proteins[0]
                if higher_or_lower == datastore.DataStore.HIGHER_PSM_SCORE:
                    if (
                        set(old_lead.peptides).issubset(set(new_lead.peptides))
                        and new_lead.identifier not in lead_protein_set
                        and old_lead.score <= new_lead.score
                        and len(old_lead.peptides) < len(new_lead.peptides)
                    ):
                        logger.info(
                            "protein {} will replace protein {} as lead, with index {}, New Num Peptides: {}, "
                            "Old Num Peptides: {}".format(
                                str(new_lead.identifier),
                                str(old_lead.identifier),
                                str(j),
                                str(len(new_lead.peptides)),
                                str(len(old_lead.peptides)),
                            )
                        )
                        lead_protein_set.add(new_lead.identifier)
                        lead_protein_set.remove(old_lead.identifier)
                        # Swap their positions in the list
                        (
                            protein_group_objects[i].proteins[0],
                            protein_group_objects[i].proteins[j],
                        ) = (new_lead, old_lead)
                        break

                if higher_or_lower == datastore.DataStore.LOWER_PSM_SCORE:
                    if (
                        set(old_lead.peptides).issubset(set(new_lead.peptides))
                        and new_lead.identifier not in lead_protein_set
                        and old_lead.score >= new_lead.score
                        and len(old_lead.peptides) < len(new_lead.peptides)
                    ):
                        logger.info(
                            "protein {} will replace protein {} as lead, with index {}, New Num Peptides: {}, "
                            "Old Num Peptides: {}".format(
                                str(new_lead.identifier),
                                str(old_lead.identifier),
                                str(j),
                                str(len(new_lead.peptides)),
                                str(len(old_lead.peptides)),
                            )
                        )
                        lead_protein_set.add(new_lead.identifier)
                        lead_protein_set.remove(old_lead.identifier)
                        # Swap their positions in the list
                        (
                            protein_group_objects[i].proteins[0],
                            protein_group_objects[i].proteins[j],
                        ) = (new_lead, old_lead)
                        break

        return protein_group_objects

    def _reassign_protein_list_leads(self, grouped_protein_objects):
        """
        This internal method corrects leads that are improperly assigned in the parsimony inference method.
        This method acts on the grouped protein objects.

        Args:
            grouped_protein_objects (list): List of [Protein][pyproteininference.physical.Protein] objects.

        Returns:
            list: List of [Protein][pyproteininference.physical.Protein] objects where leads have been
                reassigned properly.


        """

        # Get the higher or lower variable
        if not self.data.high_low_better:
            higher_or_lower = self.data.higher_or_lower()
        else:
            higher_or_lower = self.data.high_low_better

        # Sometimes we have cases where:
        # protein a maps to peptides 1,2,3
        # protein b maps to peptides 1,2
        # protein c maps to a bunch of peptides and peptide 3
        # Therefore, in the model proteins a and b are equivalent in that they map to 2 peptides together - 1 and 2.
        # peptide 3 maps to a but also to c...
        # Sometimes the model (pulp) will spit out protein b as the lead... we wish to swap protein b as the lead with
        # protein a because it will likely have a better score...
        logger.info("Potentially Reassigning Proteoin List leads...")
        lead_protein_set = set([x[0].identifier for x in grouped_protein_objects])
        for i in range(len(grouped_protein_objects)):
            for j in range(1, len(grouped_protein_objects[i])):  # Loop over all sub proteins in the group...
                # if the lead proteins peptides are a subset of one of its proteins in the group, and the secondary
                # protein is not a lead protein and its score is better than the leads... and it has more peptides...
                new_lead = grouped_protein_objects[i][j]
                old_lead = grouped_protein_objects[i][0]
                if higher_or_lower == datastore.DataStore.HIGHER_PSM_SCORE:
                    if (
                        set(old_lead.peptides).issubset(set(new_lead.peptides))
                        and new_lead.identifier not in lead_protein_set
                        and old_lead.score <= new_lead.score
                        and len(old_lead.peptides) < len(new_lead.peptides)
                    ):
                        logger.info(
                            "protein {} will replace protein {} as lead, with index {}, New Num Peptides: {}, "
                            "Old Num Peptides: {}".format(
                                str(new_lead.identifier),
                                str(old_lead.identifier),
                                str(j),
                                str(len(new_lead.peptides)),
                                str(len(old_lead.peptides)),
                            )
                        )
                        lead_protein_set.add(new_lead.identifier)
                        lead_protein_set.remove(old_lead.identifier)
                        # Swap their positions in the list
                        (
                            grouped_protein_objects[i][0],
                            grouped_protein_objects[i][j],
                        ) = (new_lead, old_lead)
                        break

                if higher_or_lower == datastore.DataStore.LOWER_PSM_SCORE:
                    if (
                        set(old_lead.peptides).issubset(set(new_lead.peptides))
                        and new_lead.identifier not in lead_protein_set
                        and old_lead.score >= new_lead.score
                        and len(old_lead.peptides) < len(new_lead.peptides)
                    ):
                        logger.info(
                            "protein {} will replace protein {} as lead, with index {}, New Num Peptides: {}, "
                            "Old Num Peptides: {}".format(
                                str(new_lead.identifier),
                                str(old_lead.identifier),
                                str(j),
                                str(len(new_lead.peptides)),
                                str(len(old_lead.peptides)),
                            )
                        )
                        lead_protein_set.add(new_lead.identifier)
                        lead_protein_set.remove(old_lead.identifier)
                        # Swap their positions in the list
                        (
                            grouped_protein_objects[i][0],
                            grouped_protein_objects[i][j],
                        ) = (new_lead, old_lead)
                        break

        return grouped_protein_objects

    def _pulp_grouper(self):
        """
        This internal function uses pulp to solve the lp problem for parsimony then performs protein grouping with the
         various internal grouping functions.

        This method assigns the variables: `grouped_scored_proteins` and `protein_group_objects`.
        These are both variables of the [DataStore object][pyproteininference.datastore.DataStore] and are
        lists of [Protein][pyproteininference.physical.Protein] objects
        and [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        """

        # Here we get the peptide to protein dictionary
        pep_prot_dict = self.data.peptide_to_protein_dictionary()

        self.data.protein_to_peptide_dictionary()

        identifiers_sorted = self.data.get_sorted_identifiers(scored=True)

        # Get all the proteins that we scored and the ones picked if picker was ran...
        data_proteins = sorted([x for x in self.data.protein_peptide_dictionary.keys() if x in identifiers_sorted])
        # Get the set of peptides for each protein...
        data_peptides = [set(self.data.protein_peptide_dictionary[x]) for x in data_proteins]
        flat_peptides_in_data = set([item for sublist in data_peptides for item in sublist])

        peptide_sets = []
        # Loop over the list of peptides...
        for k in range(len(data_peptides)):
            raw_peptides = data_peptides[k]
            peptide_set = set()
            # Loop over each individual peptide per protein...
            for peps in raw_peptides:
                peptide = peps

                # Remove mods...
                new_peptide = Psm.remove_peptide_mods(peptide)
                # Add it to a temporary set...
                peptide_set.add(new_peptide)
            # Append this set to a new list...
            peptide_sets.append(peptide_set)
            # Set that proteins peptides to be the unmodified ones...
            data_peptides[k] = peptide_set

        # Get them all...
        all_peptides = [x for x in data_peptides]
        # Remove redundant sets...
        non_redundant_peptide_sets = [set(i) for i in OrderedDict.fromkeys(frozenset(item) for item in peptide_sets)]

        # Loop over  the restricted list of peptides...
        ind_list = []
        for pep_sets in non_redundant_peptide_sets:
            # Get its index in terms of the overall list...
            ind_list.append(all_peptides.index(pep_sets))

        # Get the protein based on the index
        restricted_proteins = [data_proteins[x] for x in range(len(data_peptides)) if x in ind_list]

        # Here we get the list of all proteins
        plist = []
        for peps in pep_prot_dict.keys():
            for prots in list(pep_prot_dict[peps]):
                if prots in restricted_proteins and peps in flat_peptides_in_data:
                    plist.append(prots)

        # Here we get the unique proteins
        unique_prots = list(set(plist).union())
        unique_protein_set = set(unique_prots)

        unique_prots_sorted = [x for x in identifiers_sorted if x in unique_prots]

        # Define the protein variables with a lower bound of 0 and catgeory Integer
        prots = pulp.LpVariable.dicts("prot", indices=unique_prots_sorted, lowBound=0, cat="Integer")

        # Define our Lp Problem which is to Minimize our objective function
        prob = pulp.LpProblem("Parsimony_Problem", pulp.LpMinimize)

        # Define our objective function, which is to take the sum of all of our proteins and find the minimum set.
        prob += pulp.lpSum([prots[i] for i in prots])

        # Set up our constraints. The constrains are as follows:

        # Loop over each peptide and determine the proteins it maps to...
        # Each peptide is a constraint with the proteins it maps to having to be greater than or equal to 1
        # In the case below we see that protein 3 has a unique peptide, protein 2 is redundant

        logger.info("Sorting peptides before looping")
        for peptides in sorted(list(pep_prot_dict.keys())):
            try:
                prob += (
                    pulp.lpSum([prots[i] for i in sorted(list(pep_prot_dict[peptides])) if i in unique_protein_set])
                    >= 1
                )
            except KeyError:
                logger.info("Not including protein {} in pulp model".format(pep_prot_dict[peptides]))

        prob.solve()

        scored_data = self.data.get_protein_data()
        scored_proteins = list(scored_data)
        protein_finder = [x.identifier for x in scored_proteins]

        lead_protein_objects = []
        lead_protein_identifiers = []
        for proteins in unique_prots_sorted:
            parsimony_value = pulp.value(prots[proteins])
            if proteins in protein_finder and parsimony_value == 1:
                p_ind = protein_finder.index(proteins)
                protein_object = scored_proteins[p_ind]
                lead_protein_objects.append(protein_object)
                lead_protein_identifiers.append(protein_object.identifier)
            else:
                if parsimony_value == 1:
                    # Why are some proteins not being found when we run exclusion???
                    logger.warning("Protein {} not found with protein finder...".format(proteins))
                else:
                    pass

        self.lead_protein_objects = lead_protein_objects

        grouped_proteins = self._create_protein_groups(
            all_scored_proteins=scored_data,
            lead_protein_objects=self.lead_protein_objects,
            grouping_type=self.data.parameter_file_object.grouping_type,
        )

        regrouped_proteins = self._swissprot_and_isoform_override(
            scored_data=scored_data,
            grouped_proteins=grouped_proteins,
            override_type="soft",
            isoform_override=True,
        )

        grouped_protein_objects = regrouped_proteins["grouped_protein_objects"]
        protein_group_objects = regrouped_proteins["group_objects"]

        # Get the higher or lower variable
        hl = self.data.higher_or_lower()

        logger.info("Sorting Results based on lead Protein Score")
        grouped_protein_objects = datastore.DataStore.sort_protein_objects(
            grouped_protein_objects=grouped_protein_objects, higher_or_lower=hl
        )
        protein_group_objects = datastore.DataStore.sort_protein_group_objects(
            protein_group_objects=protein_group_objects, higher_or_lower=hl
        )

        # Run lead reassignment for the group objets and protein objects
        protein_group_objects = self._reassign_protein_group_leads(
            protein_group_objects=protein_group_objects,
        )

        grouped_protein_objects = self._reassign_protein_list_leads(grouped_protein_objects=grouped_protein_objects)

        logger.info("Re Sorting Results based on lead Protein Score")
        grouped_protein_objects = datastore.DataStore.sort_protein_objects(
            grouped_protein_objects=grouped_protein_objects, higher_or_lower=hl
        )
        protein_group_objects = datastore.DataStore.sort_protein_group_objects(
            protein_group_objects=protein_group_objects, higher_or_lower=hl
        )

        self.data.grouped_scored_proteins = grouped_protein_objects
        self.data.protein_group_objects = protein_group_objects

    def infer_proteins(self):
        """
        This method performs the Parsimony inference method and uses pulp for the LP solver.

        This method assigns the variables: `grouped_scored_proteins` and `protein_group_objects`.
        These are both variables of the [DataStore object][pyproteininference.datastore.DataStore] and are
        lists of [Protein][pyproteininference.physical.Protein] objects
        and [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        """

        if self.parameter_file_object.lp_solver == self.PULP:

            self._pulp_grouper()

        else:
            raise ValueError(
                "Parsimony cannot run if lp_solver parameter value is not one of the following: {}".format(
                    ", ".join(Inference.LP_SOLVERS)
                )
            )

        # Call assign shared peptides
        self._assign_shared_peptides(shared_pep_type=self.parameter_file_object.shared_peptides)

    def _assign_shared_peptides(self, shared_pep_type="all"):

        if not self.data.grouped_scored_proteins and self.data.protein_group_objects:
            raise ValueError(
                "Grouped Protein objects could not be found. Please run 'infer_proteins' method of the Parsimony class"
            )

        if shared_pep_type == self.ALL_SHARED_PEPTIDES:
            pass

        elif shared_pep_type == self.BEST_SHARED_PEPTIDES:
            logger.info("Assigning Shared Peptides from Parsimony to the Best Scoring Protein")
            raw_peptide_tracker = set()
            peptide_tracker = set()
            for prots in self.data.grouped_scored_proteins:
                new_psms = []
                new_raw_peptides = set()
                new_peptides = set()
                lead_prot = prots[0]
                for psm in lead_prot.psms:
                    raw_pep = psm.identifier
                    pep = psm.non_flanking_peptide
                    if raw_pep not in raw_peptide_tracker:
                        new_raw_peptides.add(raw_pep)
                        raw_peptide_tracker.add(raw_pep)
                    if pep not in peptide_tracker:
                        new_peptides.add(pep)
                        new_psms.append(psm)
                        peptide_tracker.add(pep)
                lead_prot.psms = new_psms
                lead_prot.raw_peptides = new_raw_peptides
                lead_prot.peptides = new_peptides

            raw_peptide_tracker = set()
            peptide_tracker = set()
            for group in self.data.protein_group_objects:
                lead_prot = group.proteins[0]
                new_psms = []
                new_raw_peptides = set()
                new_peptides = set()
                for psm in lead_prot.psms:
                    raw_pep = psm.identifier
                    pep = psm.non_flanking_peptide
                    if raw_pep not in raw_peptide_tracker:
                        new_raw_peptides.add(raw_pep)
                        raw_peptide_tracker.add(raw_pep)
                    if pep not in peptide_tracker:
                        new_peptides.add(pep)
                        new_psms.append(psm)
                        peptide_tracker.add(pep)

                lead_prot.psms = new_psms
                lead_prot.raw_peptides = new_raw_peptides
                lead_prot.peptides = new_peptides

        else:
            pass


class FirstProtein(Inference):
    """
    FirstProtein Inference class. This class contains methods that support the initialization of a
    FirstProtein inference method.

    Attributes:
        data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
        digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].

    """

    def __init__(self, data, digest):
        """
        FirstProtein Inference initialization method.

        Args:
            data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].

        Returns:
            object:
        """
        self.data = data
        self.digest = digest
        self.data._validate_scored_proteins()
        self.scored_data = self.data.get_protein_data()
        self.data = data

    def infer_proteins(self):
        """
        This method performs the First Protein inference method.

        This method assigns the variables: `grouped_scored_proteins` and `protein_group_objects`.
        These are both variables of the [DataStore object][pyproteininference.datastore.DataStore] and are
        lists of [Protein][pyproteininference.physical.Protein] objects
        and [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        """

        grouped_proteins = self._create_protein_groups(scored_proteins=self.scored_data)

        # Get the higher or lower variable
        hl = self.data.higher_or_lower()

        logger.info("Applying Group ID's for the First Protein Method")
        regrouped_proteins = self._apply_protein_group_ids(
            grouped_protein_objects=grouped_proteins,
        )

        grouped_protein_objects = regrouped_proteins["grouped_protein_objects"]
        protein_group_objects = regrouped_proteins["group_objects"]

        logger.info("Sorting Results based on lead Protein Score")
        grouped_protein_objects = datastore.DataStore.sort_protein_objects(
            grouped_protein_objects=grouped_protein_objects, higher_or_lower=hl
        )
        protein_group_objects = datastore.DataStore.sort_protein_group_objects(
            protein_group_objects=protein_group_objects, higher_or_lower=hl
        )

        self.data.grouped_scored_proteins = grouped_protein_objects
        self.data.protein_group_objects = protein_group_objects


class PeptideCentric(Inference):
    """
    PeptideCentric Inference class. This class contains methods that support the initialization of a
    PeptideCentric inference method.

    Attributes:
        data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
        digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].

    """

    def __init__(self, data, digest):
        """
        PeptideCentric Inference initialization method.

        Args:
            data (DataStore): [DataStore Object][pyproteininference.datastore.DataStore].
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].

        Returns:
            object:
        """
        self.data = data
        self.digest = digest
        self.data._validate_scored_proteins()
        self.scored_data = self.data.get_protein_data()

    def infer_proteins(self):
        """
        This method performs the Peptide Centric inference method.

        This method assigns the variables: `grouped_scored_proteins` and `protein_group_objects`.
        These are both variables of the [DataStore object][pyproteininference.datastore.DataStore] and are
        lists of [Protein][pyproteininference.physical.Protein] objects
        and [ProteinGroup][pyproteininference.physical.ProteinGroup] objects.

        Returns:
            None:

        """

        # Get the higher or lower variable
        hl = self.data.higher_or_lower()

        logger.info("Applying Group ID's for the Peptide Centric Method")
        regrouped_proteins = self._apply_protein_group_ids()

        grouped_protein_objects = regrouped_proteins["grouped_protein_objects"]
        protein_group_objects = regrouped_proteins["group_objects"]

        logger.info("Sorting Results based on lead Protein Score")
        grouped_protein_objects = datastore.DataStore.sort_protein_objects(
            grouped_protein_objects=grouped_protein_objects, higher_or_lower=hl
        )
        protein_group_objects = datastore.DataStore.sort_protein_group_objects(
            protein_group_objects=protein_group_objects, higher_or_lower=hl
        )

        self.data.grouped_scored_proteins = grouped_protein_objects
        self.data.protein_group_objects = protein_group_objects

    def _apply_protein_group_ids(self):
        """
        This method creates the ProteinGroup objects for the peptide_centric inference based on protein groups
        from [._create_protein_groups][pyproteininference.inference.Inference._create_protein_groups].

        Returns:
            dict: a Dictionary that contains a list of [ProteinGroup]]pyproteininference.physical.ProteinGroup]
            objects (key:"group_objects") and a list of grouped [Protein]]pyproteininference.physical.Protein]
            objects (key:"grouped_protein_objects").

        """

        grouped_protein_objects = self.data.get_protein_data()

        # Here we create group ID's
        group_id = 0
        list_of_proteins_grouped = []
        protein_group_objects = []
        for protein_group in grouped_protein_objects:
            protein_group.peptides = set(
                [Psm.split_peptide(peptide_string=x) for x in list(protein_group.raw_peptides)]
            )
            protein_list = []
            group_id = group_id + 1
            pg = ProteinGroup(group_id)
            logger.debug("Created Protein Group with ID: {}".format(str(group_id)))
            # The following loop assigns group_id's, reviewed/unreviewed status, and number of unique peptides...
            if group_id not in protein_group.group_identification:
                protein_group.group_identification.add(group_id)
            protein_group.num_peptides = len(protein_group.peptides)
            # Here append the number of unique peptides... so we can use this as secondary sorting...
            protein_list.append(protein_group)
            # Sorted protein_groups then becomes a list of lists... of protein objects

            pg.proteins = protein_list
            protein_group_objects.append(pg)
            list_of_proteins_grouped.append([protein_group])

        return_dict = {
            "grouped_protein_objects": list_of_proteins_grouped,
            "group_objects": protein_group_objects,
        }

        return return_dict



from pyproteininference import datastore  # noqa F401
from pyproteininference import export  # noqa F401
from pyproteininference import heuristic  # noqa F401
from pyproteininference import inference  # noqa F401
from pyproteininference import parameters  # noqa F401
from pyproteininference import physical  # noqa F401
from pyproteininference import pipeline  # noqa F401
from pyproteininference import reader  # noqa F401
from pyproteininference import scoring  # noqa F401



from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_peptide_centric.yaml")

TARGET_FILE_ADDITIVE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target_additive.txt")
DECOY_FILE_ADDITIVE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy_additive.txt")

TARGET_FILE_MULTIPLICATIVE = resource_filename(
    "pyproteininference", "../tests/data/test_perc_data_target_multiplicative.txt"
)
DECOY_FILE_MULTIPLICATIVE = resource_filename(
    "pyproteininference", "../tests/data/test_perc_data_decoy_multiplicative.txt"
)

PARAMETER_FILE_ADDITIVE = resource_filename(
    "pyproteininference", "../tests/data/test_params_additive_custom_score.yaml"
)
PARAMETER_FILE_MULTIPLICATIVE = resource_filename(
    "pyproteininference", "../tests/data/test_params_multiplicative_custom_score.yaml"
)


class TestReader(TestCase):
    def test_generic_and_percolator_readers(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        pep_and_prot_data_perc = pyproteininference.reader.PercolatorReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
        )
        pep_and_prot_data_perc.read_psms()

        self.assertEqual(len(pep_and_prot_data_perc.psms), 27)

        # Test to see if the values between the generic reader and the Percolator reader are identical
        for k in range(len(pep_and_prot_data.psms)):
            cur_generic = pep_and_prot_data.psms[k]
            cur_perc = pep_and_prot_data_perc.psms[k]
            self.assertEqual(cur_generic.identifier, cur_perc.identifier)
            self.assertEqual(cur_generic.pepvalue, cur_perc.pepvalue)
            self.assertEqual(cur_generic.percscore, cur_perc.percscore)
            self.assertEqual(cur_generic.possible_proteins, cur_perc.possible_proteins)
            self.assertEqual(cur_generic.psm_id, cur_perc.psm_id)
            self.assertEqual(cur_generic.qvalue, cur_perc.qvalue)

        # Test to see if we can read a list of files
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=[TARGET_FILE, TARGET_FILE],
            decoy_file=[DECOY_FILE, DECOY_FILE],
            parameter_file_object=protein_inference_parameters,
            digest=digest,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 54)

        pep_and_prot_data_perc = pyproteininference.reader.PercolatorReader(
            target_file=[TARGET_FILE, TARGET_FILE],
            decoy_file=[DECOY_FILE, DECOY_FILE],
            parameter_file_object=protein_inference_parameters,
            digest=digest,
        )
        pep_and_prot_data_perc.read_psms()

        self.assertEqual(len(pep_and_prot_data_perc.psms), 54)

        # Test to see if the values between the generic reader and the Percolator reader are identical
        for k in range(len(pep_and_prot_data.psms)):
            cur_generic = pep_and_prot_data.psms[k]
            cur_perc = pep_and_prot_data_perc.psms[k]
            self.assertEqual(cur_generic.identifier, cur_perc.identifier)
            self.assertEqual(cur_generic.pepvalue, cur_perc.pepvalue)
            self.assertEqual(cur_generic.percscore, cur_perc.percscore)
            self.assertEqual(cur_generic.possible_proteins, cur_perc.possible_proteins)
            self.assertEqual(cur_generic.psm_id, cur_perc.psm_id)
            self.assertEqual(cur_generic.qvalue, cur_perc.qvalue)

    def test_generic_functionality(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters_add = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE_ADDITIVE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters_add.digest_type,
            missed_cleavages=protein_inference_parameters_add.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters_add.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters_add.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data_add = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE_ADDITIVE,
            decoy_file=DECOY_FILE_ADDITIVE,
            parameter_file_object=protein_inference_parameters_add,
            digest=digest,
        )
        pep_and_prot_data_add.read_psms()

        self.assertEqual(len(pep_and_prot_data_add.psms), 27)

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters_mult = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE_MULTIPLICATIVE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters_mult.digest_type,
            missed_cleavages=protein_inference_parameters_mult.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters_mult.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters_mult.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data_mult = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE_MULTIPLICATIVE,
            decoy_file=DECOY_FILE_MULTIPLICATIVE,
            parameter_file_object=protein_inference_parameters_mult,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data_mult.read_psms()

        self.assertEqual(len(pep_and_prot_data_mult.psms), 27)



from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_inclusion.yaml")
COMBINED_FILE = resource_filename("pyproteininference", "../tests/data/test_data_many_alternative_proteins.txt")


class TestAltProteinFomInput(TestCase):
    def test_read_alt_proteins(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=None,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        reader = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        reader.read_psms()

        expected_result = [
            ['RAF1_HUMAN|P04049'],
            ['RAF1_HUMAN|P04049'],
            ['RAF1_HUMAN|P04049', 'ARAF_HUMAN|P10398', 'BRAF_HUMAN|P15056'],
            ['RAF1_HUMAN|P04049', 'ARAF_HUMAN|P10398'],
            ['RAF1_HUMAN|P04049', 'BRAF_HUMAN|P15056'],
            ['ARAF_HUMAN|P10398', 'BRAF_HUMAN|P15056'],
            ['ARAF_HUMAN|P10398'],
            ['RAF1_HUMAN|P04049'],
            ['TCAF1_HUMAN|Q9Y4C2'],
            ['TCAF1_HUMAN|Q9Y4C2'],
            ['B3KX72_HUMAN|B3KX72', 'HNRPU_HUMAN|Q00839'],
            ['B3KX72_HUMAN|B3KX72', 'Q96BA7_HUMAN|Q96BA7', 'HNRPU_HUMAN|Q00839'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['RPOC_SHIF8|Q0SY12'],
            ['##TCAF1_HUMAN|##Q9Y4C2'],
            ['##TCAF2_HUMAN|##A6NFQ2'],
            ['RAF1_HUMAN|P04049'],
        ]

        for i in range(len(reader.psms)):
            self.assertSetEqual(set(reader.psms[i].possible_proteins), set(expected_result[i]))

        self.assertEqual(len(reader.psms), 27)

    def test_read_alt_proteins_over_maximum(self):

        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        digest = in_silico_digest.PyteomicsDigest(
            database_path=None,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )

        # Try with generic reader

        reader = pyproteininference.reader.GenericReader(
            combined_files=COMBINED_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=True,
        )
        reader.read_psms()

        self.assertEqual(reader.MAX_ALLOWED_ALTERNATIVE_PROTEINS, 50)

        protein_list1 = [
            'Protein1',
            'Protein10',
            'Protein11',
            'Protein12',
            'Protein13',
            'Protein14',
            'Protein15',
            'Protein16',
            'Protein17',
            'Protein18',
            'Protein19',
            'Protein2',
            'Protein20',
            'Protein21',
            'Protein22',
            'Protein23',
            'Protein24',
            'Protein25',
            'Protein26',
            'Protein27',
            'Protein28',
            'Protein29',
            'Protein3',
            'Protein30',
            'Protein31',
            'Protein32',
            'Protein33',
            'Protein34',
            'Protein35',
            'Protein36',
            'Protein37',
            'Protein38',
            'Protein39',
            'Protein4',
            'Protein40',
            'Protein41',
            'Protein42',
            'Protein43',
            'Protein44',
            'Protein45',
            'Protein46',
            'Protein47',
            'Protein48',
            'Protein49',
            'Protein5',
            'Protein50',
            'Protein51',
            'Protein52',
            'Protein53',
            'Protein54',
        ]

        protein_list2 = [
            '##Protein1',
            '##Protein10',
            '##Protein11',
            '##Protein12',
            '##Protein13',
            '##Protein14',
            '##Protein15',
            '##Protein16',
            '##Protein17',
            '##Protein18',
            '##Protein19',
            '##Protein2',
            '##Protein20',
            '##Protein21',
            '##Protein22',
            '##Protein23',
            '##Protein24',
            '##Protein25',
            '##Protein26',
            '##Protein27',
            '##Protein28',
            '##Protein29',
            '##Protein3',
            '##Protein30',
            '##Protein31',
            '##Protein32',
            '##Protein33',
            '##Protein34',
            '##Protein35',
            '##Protein36',
            '##Protein37',
            '##Protein38',
            '##Protein39',
            '##Protein4',
            '##Protein40',
            '##Protein41',
            '##Protein42',
            '##Protein43',
            '##Protein44',
            '##Protein45',
            '##Protein46',
            '##Protein47',
            '##Protein48',
            '##Protein49',
            '##Protein5',
            '##Protein50',
            '##Protein51',
            '##Protein52',
            '##Protein53',
            '##Protein54',
        ]

        self.assertListEqual(
            reader.psms[0].possible_proteins,
            protein_list1,
        )
        self.assertListEqual(
            reader.psms[1].possible_proteins,
            protein_list2,
        )

        percdigest = in_silico_digest.PyteomicsDigest(
            database_path=None,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )

        # Try with percolator reader
        percreader = pyproteininference.reader.PercolatorReader(
            combined_files=COMBINED_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=percdigest,
            append_alt_from_db=True,
        )

        percreader.read_psms()

        self.assertEqual(percreader.MAX_ALLOWED_ALTERNATIVE_PROTEINS, 50)

        self.assertListEqual(
            percreader.psms[0].possible_proteins,
            protein_list1,
        )
        self.assertListEqual(
            percreader.psms[1].possible_proteins,
            protein_list2,
        )



import logging
import math
import sys
from functools import reduce

import numpy

logger = logging.getLogger(__name__)

# set up our logger
logging.basicConfig(
    stream=sys.stderr,
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class Score(object):
    """
    Score class that contains methods to do a variety of scoring methods on the
    [Psm][pyproteininference.physical.Psm] objects
    contained inside of [Protein][pyproteininference.physical.Protein] objects.

    Methods in the class loop over each Protein object and creates a protein "score" variable using the Psm object
    scores.

    Methods score all proteins from `scoring_input` from [DataStore object][pyproteininference.datastore.DataStore].
    The PSM score that is used is determined from
    [create_scoring_input][pyproteininference.datastore.DataStore.create_scoring_input].

    Each scoring method will set the following attributes for
    the [DataStore object][pyproteininference.datastore.DataStore].

    1. `score_method`; This is the full name of the score method.
    2. `short_score_method`; This is the short name of the score method.
    3. `scored_proteins`; This is a list of [Protein][pyproteininference.physical.Protein] objects
    that have been scored.

    Attributes:
        pre_score_data (list): This is a list of [Protein][pyproteininference.physical.Protein] objects
            that contain [Psm][pyproteininference.physical.Psm] objects.
        data (DataStore): [DataStore][pyproteininference.datastore.DataStore] object.

    """

    BEST_PEPTIDE_PER_PROTEIN = "best_peptide_per_protein"
    ITERATIVE_DOWNWEIGHTED_LOG = "iterative_downweighted_log"
    MULTIPLICATIVE_LOG = "multiplicative_log"
    DOWNWEIGHTED_MULTIPLICATIVE_LOG = "downweighted_multiplicative_log"
    DOWNWEIGHTED_VERSION2 = "downweighted_version2"
    TOP_TWO_COMBINED = "top_two_combined"
    GEOMETRIC_MEAN = "geometric_mean"
    ADDITIVE = "additive"

    SCORE_METHODS = [
        BEST_PEPTIDE_PER_PROTEIN,
        ITERATIVE_DOWNWEIGHTED_LOG,
        MULTIPLICATIVE_LOG,
        DOWNWEIGHTED_MULTIPLICATIVE_LOG,
        DOWNWEIGHTED_VERSION2,
        TOP_TWO_COMBINED,
        GEOMETRIC_MEAN,
        ADDITIVE,
    ]

    SHORT_BEST_PEPTIDE_PER_PROTEIN = "bppp"
    SHORT_ITERATIVE_DOWNWEIGHTED_LOG = "idwl"
    SHORT_MULTIPLICATIVE_LOG = "ml"
    SHORT_DOWNWEIGHTED_MULTIPLICATIVE_LOG = "dwml"
    SHORT_DOWNWEIGHTED_VERSION2 = "dw2"
    SHORT_TOP_TWO_COMBINED = "ttc"
    SHORT_GEOMETRIC_MEAN = "gm"
    SHORT_ADDITIVE = "add"

    SHORT_SCORE_METHODS = [
        SHORT_BEST_PEPTIDE_PER_PROTEIN,
        SHORT_ITERATIVE_DOWNWEIGHTED_LOG,
        SHORT_MULTIPLICATIVE_LOG,
        SHORT_DOWNWEIGHTED_MULTIPLICATIVE_LOG,
        SHORT_DOWNWEIGHTED_VERSION2,
        SHORT_TOP_TWO_COMBINED,
        SHORT_GEOMETRIC_MEAN,
        SHORT_ADDITIVE,
    ]

    MULTIPLICATIVE_SCORE_TYPE = "multiplicative"
    ADDITIVE_SCORE_TYPE = "additive"

    SCORE_TYPES = [MULTIPLICATIVE_SCORE_TYPE, ADDITIVE_SCORE_TYPE]

    def __init__(self, data):
        """
        Initialization method for the Score class.

        Args:
            data (DataStore): [DataStore][pyproteininference.datastore.DataStore] object.

        Raises:
            ValueError: If the variable `scoring_input` for the [DataStore][pyproteininference.datastore.DataStore]
                object is Empty "[]" or does not exist "None".

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
        """
        if data.scoring_input:
            self.pre_score_data = data.scoring_input
        else:
            raise ValueError(
                "scoring input not found in data object - Please run 'create_scoring_input' method from "
                "DataStore to run any scoring type"
            )
        self.data = data

    def score_psms(self, score_method="multiplicative_log"):
        """
        This method dispatches to the actual scoring method given a string input that is defined in the
        [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter] object.

        Args:
            score_method (str): This is a string that represents which scoring method to call.

        Raises:
            ValueError: Will Error out if the score_method is not present in the constant `SCORE_METHODS`.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.score_psms(score_method="best_peptide_per_protein")
        """

        if score_method not in self.SCORE_METHODS:
            raise ValueError(
                "score method '{}' is not a proper method. Score method must be one of the following: '{}'".format(
                    score_method, ", ".join(self.SCORE_METHODS)
                )
            )
        else:
            if score_method == self.BEST_PEPTIDE_PER_PROTEIN:
                self.best_peptide_per_protein()
            if score_method == self.ITERATIVE_DOWNWEIGHTED_LOG:
                self.iterative_down_weighted_log()
            if score_method == self.MULTIPLICATIVE_LOG:
                self.multiplicative_log()
            if score_method == self.DOWNWEIGHTED_MULTIPLICATIVE_LOG:
                self.down_weighted_multiplicative_log()
            if score_method == self.DOWNWEIGHTED_VERSION2:
                self.down_weighted_v2()
            if score_method == self.TOP_TWO_COMBINED:
                self.top_two_combied()
            if score_method == self.GEOMETRIC_MEAN:
                self.geometric_mean_log()
            if score_method == self.ADDITIVE:
                self.additive()

    def best_peptide_per_protein(self):
        """
        This method uses a best peptide per protein scoring scheme.
        The top scoring Psm for each protein is selected as the overall Protein object score.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.best_peptide_per_protein()

        """

        all_scores = []

        logger.info("Scoring Proteins with BPPP")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()
            score = min([float(x) for x in val_list])

            protein.score = score

            all_scores.append(protein)
        # Here do ascending sorting because a lower pep or q value is better
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=False)

        self.data.protein_score = self.BEST_PEPTIDE_PER_PROTEIN
        self.data.short_protein_score = self.SHORT_BEST_PEPTIDE_PER_PROTEIN
        self.data.scored_proteins = all_scores

    def fishers_method(self):
        """
        This method uses a fishers method scoring scheme.
\
        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.fishers_method()

         """

        all_scores = []
        logger.info("Scoring Proteins with fishers method")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()
            score = -2 * sum([math.log(x) for x in val_list])

            protein.score = score

            all_scores.append(protein)
        # Here reverse the sorting to descending because a higher score is better
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)
        self.data.protein_score = "fishers_method"
        self.data.short_protein_score = "fm"
        self.data.scored_proteins = all_scores

    def multiplicative_log(self):
        """
        This method uses a Multiplicative Log scoring scheme.
        The selected Psm score from all the peptides per protein are multiplied together and we take -Log(X)
        of the multiplied Peptide scores.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.multiplicative_log()
        """

        # Instead of making all_scores a list... make it a generator??

        all_scores = []
        logger.info("Scoring Proteins with Multiplicative Log Method")
        for protein in self.pre_score_data:
            # We create a generator of val_list...
            val_list = protein.get_psm_scores()

            combine = reduce(lambda x, y: x * y, val_list)
            if combine == 0:
                combine = sys.float_info.min
            score = -math.log(combine)
            protein.score = score

            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = self.MULTIPLICATIVE_LOG
        self.data.short_protein_score = self.SHORT_MULTIPLICATIVE_LOG
        self.data.scored_proteins = all_scores

    def down_weighted_multiplicative_log(self):
        """
        This method uses a Multiplicative Log scoring scheme.
        The selected PSM score from all the peptides per protein are multiplied together and
        then this number is divided by the set PSM scores mean raised to the number of peptides for that protein
        then we take -Log(X) of the following value.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.down_weighted_multiplicative_log()
        """

        score_list = []
        for proteins in self.pre_score_data:
            cur_scores = proteins.get_psm_scores()
            for scores in cur_scores:
                score_list.append(scores)
        score_mean = numpy.mean(score_list)

        all_scores = []
        logger.info("Scoring Proteins with DWML method")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()
            # Divide by the score mean raised to the length of the number of unique peptides for the protein
            # This is an attempt to normalize for number of peptides per protein
            combine = reduce(lambda x, y: x * y, val_list)
            if combine == 0:
                combine = sys.float_info.min
            score = -math.log(combine / (score_mean ** len(val_list)))
            protein.score = score

            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = self.DOWNWEIGHTED_MULTIPLICATIVE_LOG
        self.data.short_protein_score = self.SHORT_DOWNWEIGHTED_MULTIPLICATIVE_LOG
        self.data.scored_proteins = all_scores

    def top_two_combied(self):
        """
        This method uses a Top Two scoring scheme.
        The top two scores for each protein are multiplied together and we take -Log(X) of the multiplied value.
        If a protein only has 1 score/peptide, then we only do -Log(X) of the 1 peptide score.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.top_two_combied()
        """

        all_scores = []
        logger.info("Scoring Proteins with Top Two Method")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()

            try:
                # Try to combine the top two scores
                # Divide by 2 to attempt to normalize the value
                score = -math.log((val_list[0] * val_list[1]) / 2)
            except IndexError:
                # If there is only 1 score/1 peptide then just use the 1 peptide provided
                score = -math.log(val_list[0])

            protein.score = score
            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = self.TOP_TWO_COMBINED
        self.data.short_protein_score = self.SHORT_TOP_TWO_COMBINED
        self.data.scored_proteins = all_scores

    def down_weighted_v2(self):
        """
        This method uses a Downweighted Multiplicative Log scoring scheme.
        Each peptide is iteratively downweighted by raising the peptide QValue or PepValue to the
        following power (1/(1+index_number)).
        Where index_number is the peptide number per protein.
        Each score for a protein provides less and less weight iteratively.

        We also take -Log(X) of the final score here.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.down_weighted_v2()
        """

        all_scores = []
        logger.info("Scoring Proteins with down weighted v2 method")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()

            # Here take each score and raise it to the power of (1/(1+index_number)).
            # This downweights each successive score by reducing its weight in a decreasing fashion
            # Basically, each score for a protein will provide less and less weight iteratively
            val_list = [val_list[x] ** (1 / float(1 + x)) for x in range(len(val_list))]
            # val_list = [val_list[x]**(1/float(1+(float(x)/10))) for x in range(len(val_list))]
            score = -math.log(reduce(lambda x, y: x * y, val_list))

            protein.score = score
            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = self.DOWNWEIGHTED_VERSION2
        self.data.short_protein_score = self.SHORT_DOWNWEIGHTED_VERSION2
        self.data.scored_proteins = all_scores

    def iterative_down_weighted_log(self):
        """
        This method uses a Downweighted Multiplicative Log scoring scheme.
        Each peptide is iteratively downweighted by multiplying the peptide QValue or PepValue to
        the following (1+index_number).
        Where index_number is the peptide number per protein.
        Each score for a protein provides less and less weight iteratively.

        We also take -Log(X) of the final score here.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.iterative_down_weighted_log()
        """

        all_scores = []
        logger.info("Scoring Proteins with IDWL method")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()

            # Here take each score and multiply it by its index number).
            # This downweights each successive score by reducing its weight in a decreasing fashion
            # Basically, each score for a protein will provide less and less weight iteratively
            val_list = [val_list[x] * (float(1 + x)) for x in range(len(val_list))]
            # val_list = [val_list[x]**(1/float(1+(float(x)/10))) for x in range(len(val_list))]
            combine = reduce(lambda x, y: x * y, val_list)
            if combine == 0:
                combine = sys.float_info.min
            score = -math.log(combine)
            protein.score = score

            protein.score = score
            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = self.ITERATIVE_DOWNWEIGHTED_LOG
        self.data.short_protein_score = self.SHORT_ITERATIVE_DOWNWEIGHTED_LOG
        self.data.scored_proteins = all_scores

    def geometric_mean_log(self):
        """
        This method uses a Geometric Mean scoring scheme.

        We also take -Log(X) of the final score here.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.geometric_mean_log()
        """

        all_scores = []
        logger.info("Scoring Proteins. with GML method")
        for protein in self.pre_score_data:
            psm_scores = protein.get_psm_scores()
            val_list = []
            for vals in psm_scores:
                val_list.append(float(vals))
                combine = reduce(lambda x, y: x * y, val_list)
                if combine == 0:
                    combine = sys.float_info.min
                pre_log_score = combine ** (1 / float(len(val_list)))
            score = -math.log(pre_log_score)

            protein.score = score
            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = self.GEOMETRIC_MEAN
        self.data.short_protein_score = self.SHORT_GEOMETRIC_MEAN
        self.data.scored_proteins = all_scores

    def iterative_down_weighted_v2(self):
        """
        The following method is an experimental method essentially used for future development of potential scoring
        schemes.
        """

        all_scores = []
        logger.info("Scoring Proteins with iterative down weighted v2 method")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()

            # Here take each score and raise it to the power of (1/(1+index_number)).
            # This downweights each successive score by reducing its weight in a decreasing fashion
            # Basically, each score for a protein will provide less and less weight iteratively
            val_list = [val_list[x] ** (1 / float(1 + x)) for x in range(len(val_list))]
            # val_list = [val_list[x]**(1/float(1+(float(x)/10))) for x in range(len(val_list))]
            score = -math.log(reduce(lambda x, y: x * y, val_list))

            protein.score = score
            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = "iterative_downweighting2"
        self.data.short_protein_score = "idw2"
        self.data.scored_proteins = all_scores

    def additive(self):
        """
        This method uses an additive scoring scheme.
        The method can only be used if a larger PSM score is a better PSM score such as the Percolator score.

        Examples:
            >>> score = pyproteininference.scoring.Score(data=data)
            >>> score.additive()
        """

        all_scores = []
        logger.info("Scoring Proteins with additive method")
        for protein in self.pre_score_data:
            val_list = protein.get_psm_scores()

            # Take the sum of our scores
            score = sum(val_list)

            protein.score = score
            all_scores.append(protein)

        # Higher score is better as a smaller q or pep in a -log will give a larger value
        all_scores = sorted(all_scores, key=lambda k: k.score, reverse=True)

        self.data.protein_score = self.ADDITIVE
        self.data.short_protein_score = self.SHORT_ADDITIVE
        self.data.scored_proteins = all_scores
